$: << File.dirname(__FILE__)
require 'rake_helpers'

namespace :nodeclass do
  desc 'List node classes'
  repeatable_task :list, [:match] => :environment do |task|
    names = NodeClass.all.map(&:name)
    regex = task.get_parameter(:match) # if nil, everything matches
    names.grep(/#{regex}/).map{|n| puts n}
  end

  desc 'Add a new node class'
  repeatable_task :add, [:name, :onexists] => :environment do |task|
    unless name = task.get_parameter(:name)
      puts 'Must specify class name (name=<class>).'
      exit 1
    end

    if NodeClass.find_by_name(name)
      if task.should_fail_on_exists
        puts 'Class already exists!'
        exit 1
      else
        puts 'Class already exists, skipping.'
        next
      end
    end

    NodeClass.new(:name => name).save!
    puts 'Class successfully created!'
  end

  desc 'Delete a node class'
  repeatable_task :del, [:name] => :environment do |task|
    begin
      get_class(task.get_parameter(:name)).destroy
      puts 'Group successfully deleted!'
    rescue NoMethodError
      puts 'Class doesn\'t exist!'
      exit 1
    rescue => e
      puts e.message
      exit 1
    end
  end
end

namespace :nodegroup do
  desc 'List node groups'
  repeatable_task :list, [:match] => :environment do |task|
    names = NodeGroup.all.map(&:name)
    regex = task.get_parameter(:match) # if nil, everything matches
    names.grep(/#{regex}/).map{|n| puts n}
  end

  desc 'List classes that belong to a node group'
  repeatable_task :listclasses, [:name] => :environment do |task|
    nodegroup = get_group(task.get_parameter(:name))

    begin
      nodegroup.node_classes.map(&:name).map{|n| puts n}
    rescue => e
      puts e.message
      exit 1
    end
  end

  desc 'List child groups that belong to a node group'
  repeatable_task :listgroups, [:name] => :environment do |task|
    nodegroup = get_group(task.get_parameter(:name))

    begin
      nodegroup.node_groups.map(&:name).map{|n| puts n}
    rescue => e
      puts e.message
      exit 1
    end
  end

  desc 'Add a new node group'
  repeatable_task :add, [:name, :onexists] => :environment do |task|
    unless name = task.get_parameter(:name)
      puts 'Must specify group name (name=<group>).'
      exit 1
    end

    if NodeGroup.find_by_name(name)
      if task.should_fail_on_exists
        puts 'Group already exists!'
        exit 1
      else
        puts 'Group already exists, skipping.'
        next
      end
    end

    nodegroup = NodeGroup.new(:name => name)

    if ENV['classes']
      classes = ENV['classes'].split(',').map(&:strip)
      nodegroup.node_classes = NodeClass.find_all_by_name(classes)
    end

    nodegroup.save!
    puts 'Group successfully created!'
  end

  desc 'Add a class to a nodegroup'
  repeatable_task :addclass, [:name, :class] => :environment do |task|
    nodegroup = get_group(task.get_parameter(:name))
    nodeclass = get_class(task.get_parameter(:class))

    begin
      classes = nodegroup.node_classes
      if classes.include?(nodeclass)
        puts "Group '#{nodegroup.name}' already includes class '#{nodeclass.name}'"
      else
        classes << nodeclass
        puts "Class '#{nodeclass.name}' added to node group '#{nodegroup.name}'"
      end
    rescue => e
      puts e.message
      exit 1
    end
  end

  desc 'Add a child group to a nodegroup'
  repeatable_task :addgroup, [:name, :group] => :environment do |task|
    nodegroup = get_group(task.get_parameter(:name))
    addlgroup = get_group(task.get_parameter(:group))

    begin
      groups = nodegroup.node_groups
      if groups.include?(addlgroup)
        puts "Group '#{nodegroup.name}' already includes group '#{addlgroup.name}'"
      else
        groups << addlgroup
        puts "Group '#{addlgroup.name}' added to node group '#{nodegroup.name}'"
      end
    rescue => e
      puts e.message
      exit 1
    end
  end

  desc 'Remove a class from a nodegroup'
  repeatable_task :delclass, [:name, :class] => :environment do |task|
    nodegroup = get_group(task.get_parameter(:name))
    nodeclass = get_class(task.get_parameter(:class))

    begin
      classes = nodegroup.node_classes
      unless classes.include?(nodeclass)
        puts "Group '#{nodegroup.name}' does not include class '#{nodeclass.name}'"
      else
        classes.delete(nodeclass)
        puts "Class '#{nodeclass.name}' removed from node group '#{nodegroup.name}'"
      end
    rescue => e
      puts e.message
      exit 1
    end
  end

  desc 'Remove a child group from a nodegroup'
  repeatable_task :delgroup, [:name, :group] => :environment do |task|
    nodegroup = get_group(task.get_parameter(:name))
    addlgroup = get_group(task.get_parameter(:group))

    begin
      groups = nodegroup.node_groups
      unless groups.include?(addlgroup)
        puts "Group '#{nodegroup.name}' does not include groups '#{addlgroup.name}'"
      else
        groups.delete(addlgroup)
        puts "Group '#{addlgroup.name}' removed from node group '#{nodegroup.name}'"
      end
    rescue => e
      puts e.message
      exit 1
    end
  end

  # deprecated - use variables instead
  repeatable_task :parameters, [:name, :parameters] => :environment do |task|
    puts "The nodegroup:parameters task is deprecated. Please use the nodegroup:variables task instead"

    Rake::Task["nodegroup:variables"].invoke(task.get_parameter(:name), task.get_parameter(:parameters))
  end

  desc 'Show/Edit/Add variables for a node group'
  repeatable_task :variables, [:name, :variables] => :environment do |task|
    nodegroup = get_group(task.get_parameter(:name))

    # Show variables
    unless variables = task.get_parameter(:variables)
      nodegroup.parameters.each do |p|
        puts "#{p.key}=#{p.value}"
      end
      next
    end

    # Split on ',', do not split on escaped commas (i.e. '\,')
    given_parameters = Hash[ variables.split(/(?<!\\),/).map do |param|
      # Split on '=', do not split on escaped equal signs (i.e. '\=')
      param_array = param.split(/(?<!\\)=/,2).map do |element|
        # Remove slashes so that equal signs and commas are no longer escaped
        element.gsub('\\=', '=').gsub('\\,', ',')
      end

      if param_array.size != 2
        raise ArgumentError, "Could not parse variable #{param_array.first} given. Perhaps you're missing a '='"
      end
      if param_array[0].nil? or param_array[0].empty?
        raise ArgumentError, "Could not parse variables. Please check your format. Perhaps you need to name a variable before a '='"
      end
      if param_array[1].nil? or param_array[1].empty?
        raise ArgumentError, "Could not parse variables #{param_array.first}. Please check your format"
      end
      param_array
    end ]

    begin
      ActiveRecord::Base.transaction do
        given_parameters.each do |key, value|
          param, *dupes = *nodegroup.parameters.find_all_by_key(key)
          if param
            # Change existing variables
            param.value = value
            param.save!
            # If there were duplicate variables from the previous buggy version of
            # this code, remove them
            dupes.each { |d| d.destroy }
          else
            # Create new variables
            nodegroup.parameters.create(:key => key, :value => value)
          end
        end

        nodegroup.save!
        puts "Node group variables successfully edited for #{nodegroup.name}!"
      end
    rescue => e
      puts "There was a problem saving the node group: #{e.message}"
      exit 1
    end
  end

  desc 'Edit a node group'
  repeatable_task :edit, [:name, :classes] => :environment do |task|
    nodegroup = get_group(task.get_parameter(:name))

    next unless classes = task.get_parameter(:classes)

    begin
      classes = classes.split(',').map(&:strip)
      nodegroup.node_classes = NodeClass.find_all_by_name(classes)
      nodegroup.save!
      puts 'Group successfully edited!'
    rescue => e
      puts e.message
      exit 1
    end
  end

  desc 'Delete a node group'
  repeatable_task :del, [:name] => :environment do |task|
    begin
      get_group(task.get_parameter(:name)).destroy
      puts 'Group successfully deleted!'
    rescue => e
      puts e.message
      exit 1
    end
  end

  desc 'Automatically adds all nodes to a group'
  repeatable_task :add_all_nodes, [:name] => :environment do |task|
    unless ENV['group'].nil?
      puts 'WARNING: The environment variable "group" has been deprecated and may be removed in a future release.'
      puts 'Please use "name" to specify the group name instead.'
    end

    group_name = task.get_parameter(:name) || ENV['group']

    group = get_group(group_name)

    begin
      Node.all.each do |node|
        node_groups = node.node_groups
        node_groups << group unless node_groups.include?(group)
      end
    rescue => e
      puts "There was a problem adding all nodes to the group '#{group}': #{e.message}"
      exit 1
    end
  end

  desc 'Automatically adds all MCollective nodes to the MCollective group'
  repeatable_task :add_mcollective_nodes, [] => :environment do |task|
    mco_group = get_group('mcollective')
    no_mco_group = get_group('no mcollective')

    # Ensure all nodes are in the default group and one of the two MCollective
    # groups.  If they're in both MCO groups then remove it from the
    # "MCollective" group.
    begin
      Node.all.each do |node|
        node_groups = node.node_groups
        if node_groups.include?(mco_group)
          if node_groups.include?(no_mco_group)
            node_groups.delete(mco_group)
          end
        else
          unless node_groups.include?(no_mco_group)
            node_groups << mco_group
          end
        end
      end
    rescue => e
      puts "There was a problem adding all nodes to the group '#{mco_group}': #{e.message}"
      exit 1
    end
  end

  desc "Add classparam(s) to a nodegroup"
  repeatable_task :addclassparam, [:name, :class, :param, :value] => :environment do |task|
    group = get_group(task.get_parameter(:name))

    unless clazz = task.get_parameter(:class)
      puts 'Must specify class to apply parameters to'
      exit 1
    end

    unless param = task.get_parameter(:param)
      puts 'Must specify `param` to be added'
      exit 1
    end

    unless value = task.get_parameter(:value)
      puts 'Must specify `value` for parameter'
      exit 1
    end

    node_class = NodeClass.find_by_id_or_name!(clazz)
    class_membership = NodeGroupClassMembership.find_by_node_group_id_and_node_class_id(group.id, node_class.id)

    begin
      parameter = class_membership.parameters.find_by_key(param)
      if parameter
        parameter.value = value
        parameter.save!
      else
        class_membership.parameters.create(:key => param, :value => value)
        class_membership.save!
      end

      puts "Successfully set param '#{param} = \"#{value}\"' to node group '#{group.name}', class '#{node_class.name}'"
    rescue => e
      puts "There was a problem saving the node group class parameter: #{e.message}"
      exit 1
    end
  end

  desc "List classparam(s) for a nodegroup/class"
  repeatable_task :listclassparams, [:name, :class] => :environment do |task|
    group = get_group(task.get_parameter(:name))

    unless clazz = task.get_parameter(:class)
      puts 'Must specify class to list parameters for'
      exit 1
    end

    node_class = NodeClass.find_by_id_or_name!(clazz)
    class_membership = NodeGroupClassMembership.find_by_node_group_id_and_node_class_id(group.id, node_class.id)

    class_membership.parameters.each do |p|
      puts "#{p.key} = \"#{p.value}\""
    end
  end

  desc 'Remove a class param from a node group'
  repeatable_task :delclassparam, [:name, :class, :param] => :environment do |task|
    group = get_group(task.get_parameter(:name))

    unless clazz = task.get_parameter(:class)
      puts 'Must specify class to delete parameter from'
      exit 1
    end

    unless param = task.get_parameter(:param)
      puts 'Must specify `param` to be added'
      exit 1
    end

    node_class = NodeClass.find_by_id_or_name!(clazz)
    class_membership = NodeGroupClassMembership.find_by_node_group_id_and_node_class_id(group.id, node_class.id)

    begin
      parameter = class_membership.parameters.find_by_key(param)
      parameter.destroy
    rescue => e
      puts e.message
      exit 1
    end
  end
end
